module.exports = new (function() {

  var SETTINGS_DEPRECTED_VAL = './settings.json';
  var SETTINGS_JS_FILE = './nightwatch.conf.js';
  var Cli = require('./cli.js');

  this.load = function () {
    // use default nightwatch.json file if we haven't received another value

    if (Cli.command('config').isDefault(this.argv.config)) {
      var defaultValue = Cli.command('config').defaults();
      var localJsValue = path.resolve(SETTINGS_JS_FILE);

      if (fs.existsSync(SETTINGS_JS_FILE)) {
        this.argv.config = localJsValue;
      } else if (fs.existsSync(defaultValue)) {
        this.argv.config = path.join(path.resolve('./'), this.argv.config);
      } else if (fs.existsSync(SETTINGS_DEPRECTED_VAL)) {
        this.argv.config = path.join(path.resolve('./'), SETTINGS_DEPRECTED_VAL);
      } else {
        var binFolder = path.resolve(__dirname + '/../../../bin');
        var defaultFile = path.join(binFolder, this.argv.config);
        if (fs.existsSync(defaultFile)) {
          this.argv.config = defaultFile;
        } else {
          this.argv.config = path.join(binFolder, SETTINGS_DEPRECTED_VAL);
        }
      }
    } else {
      this.argv.config = path.resolve(this.argv.config);
    }

    this.argv.env = typeof this.argv.env == 'string' ? this.argv.env : 'default';

    // reading the settings file
    try {
      this.settings = require(this.argv.config);
      this.replaceEnvVariables();

      this.manageSelenium = !this.isParallelMode() && this.settings.selenium &&
        this.settings.selenium.start_process || false;

      if (typeof this.settings.src_folders == 'string') {
        this.settings.src_folders = [this.settings.src_folders];
      }

      this.settings.output = this.settings.output || typeof this.settings.output == 'undefined';
    } catch (ex) {
      Logger.error(ex);
      this.settings = {};
    }

    return this;
  }

})();



/**
 * Looks for pattern ${VAR_NAME} in settings
 * @param {Object} [target]
 */
function replaceEnvVariables(target) {
  target = target || this.settings;
  for (var key in target) {
    switch(typeof target[key]) {
      case 'object':
        this.replaceEnvVariables(target[key]);
        break;
      case 'string':
        target[key] = target[key].replace(/\$\{(\w+)\}/g, function(match, varName) {
          return process.env[varName] || '${' + varName + '}';
        });
        break;
    }
  }

  return this;
}

/**
 * @returns {CliRunner}
 */
function setOutputFolder() {
  var isDisabled = this.settings.output_folder === false;
  var isDefault = Cli.command('output').isDefault(this.argv.output);

  this.output_folder = isDisabled ? false : (isDefault && this.settings.output_folder || this.argv.output);
  return this;
}

function inheritFromDefaultEnv() {
  if (this.argv.env == 'default') {
    return;
  }
  var defaultEnv = this.settings.test_settings['default'] || {};
  for (var key in defaultEnv) {
    if (typeof this.test_settings[key] == 'undefined') {
      this.test_settings[key] = defaultEnv[key];
    }
  }
}

/**
 * Validates and parses the test settings
 * @param {object} [settings]
 * @param {function} [done]
 * @returns {CliRunner}
 */
function parseTestSettings(settings, done) {
  // checking if the env passed is valid
  if (!this.settings.test_settings) {
    throw new Error('No testing environment specified.');
  }

  var envs = this.argv.env.split(',');
  for (var i = 0; i < envs.length; i++) {
    if (!(envs[i] in this.settings.test_settings)) {
      throw new Error('Invalid testing environment specified: ' + envs[i]);
    }
  }

  if (envs.length > 1) {
    this.setupParallelMode(envs, done);
    return this;
  }

  this.initTestSettings(this.argv.env, settings);

  if (this.parallelModeWorkers()) {
    this.setupParallelMode(null, done);
  }

  return this;
},

function setGlobalOutputOptions() {
  this.test_settings.output = this.test_settings.output || (this.settings.output && typeof this.test_settings.output == 'undefined');
  this.test_settings.silent = this.test_settings.silent || typeof this.test_settings.silent == 'undefined';
  if (this.argv.verbose) {
    this.test_settings.silent = false;
  }

  return this;
}

/**
 * Sets the specific test settings for the specified environment
 * @param {string} [env]
 * @param {object} [settings]
 * @returns {CliRunner}
 */
function initTestSettings(env, settings) {
  // picking the environment specific test settings
  this.test_settings = env && this.settings.test_settings[env] || {};
  if (env) {
    this.test_settings.custom_commands_path = this.settings.custom_commands_path || '';
    this.test_settings.custom_assertions_path = this.settings.custom_assertions_path || '';
    this.test_settings.page_objects_path = this.settings.page_objects_path || '';

    this.inheritFromDefaultEnv();
    this.updateTestSettings(settings || {});
    this.readExternalGlobals();
  }

  this.setGlobalOutputOptions();

  if (typeof this.test_settings.test_workers != 'undefined') {
    this.settings.test_workers = this.test_settings.test_workers;
  } else if (typeof this.settings.test_workers != 'undefined') {
    this.test_settings.test_workers = this.settings.test_workers;
  }

  if (typeof this.test_settings.test_runner != 'undefined') {
    this.settings.test_runner = this.test_settings.test_runner;
  }

  if (typeof this.argv.skipgroup == 'string') {
    this.test_settings.skipgroup = this.argv.skipgroup.split(',');
  }

  if (this.argv.filter) {
    this.test_settings.filename_filter = this.argv.filter;
  }

  if (this.argv.tag) {
    this.test_settings.tag_filter = this.argv.tag;
  }

  if (typeof this.argv.skiptags == 'string') {
    this.test_settings.skiptags = this.argv.skiptags.split(',');
  }

  return this;
}

/**
 *
 * @param {Object} [test_settings]
 * @returns {CliRunner}
 */
 function updateTestSettings(test_settings) {
  if (this.parallelMode && !this.parallelModeWorkers()) {
    return this;
  }

  if (test_settings && typeof test_settings == 'object') {
    for (var key in test_settings) {
      this.test_settings[key] = test_settings[key];
    }
  }

  this.settings.selenium = this.settings.selenium || {};

  // overwrite selenium settings per environment
  if (Utils.isObject(this.test_settings.selenium)) {
    for (var prop in this.test_settings.selenium) {
      this.settings.selenium[prop] = this.test_settings.selenium[prop];
    }
  }

  this.manageSelenium = !this.isParallelMode() && this.settings.selenium.start_process || false;
  this.startSession = this.settings.selenium.start_session || typeof this.settings.selenium.start_session == 'undefined';

  this.mergeSeleniumOptions();
  this.disableCliColorsIfNeeded();

  this.endSessionOnFail = typeof this.settings.end_session_on_fail == 'undefined' || this.settings.end_session_on_fail;
  if (typeof this.test_settings.end_session_on_fail != 'undefined') {
    this.endSessionOnFail = this.test_settings.end_session_on_fail;
  }
  return this;
},

/**
 * Backwards compatible method which attempts to merge deprecated driver specific options for selenium
 * @returns {CliRunner}
 */
function mergeSeleniumOptions() {
  if (!this.manageSelenium) {
    return this;
  }
  this.settings.selenium.cli_args = this.settings.selenium.cli_args || {};
  this.mergeCliArgs();

  var deprecationNotice = function(propertyName, newSettingName) {
    ErrorHandler.logWarning('DEPRECATION NOTICE: Property ' + propertyName + ' is deprecated since v0.5. Please' +
      ' use the "cli_args" object on the "selenium" property to define "' + newSettingName + '". E.g.:');
    var demoObj = '{\n' +
      '  "cli_args": {\n' +
      '    "'+ Logger.colors.yellow(newSettingName) +'": "<VALUE>"\n' +
      '  }\n' +
      '}';

    console.log(demoObj, '\n');
  };

  if (this.test_settings.firefox_profile) {
    deprecationNotice('firefox_profile', 'webdriver.firefox.profile');
    this.settings.selenium.cli_args['webdriver.firefox.profile'] = this.test_settings.firefox_profile;
  }
  if (this.test_settings.chrome_driver) {
    deprecationNotice('chrome_driver', 'webdriver.chrome.driver');
    this.settings.selenium.cli_args['webdriver.chrome.driver'] = this.test_settings.chrome_driver;
  }
  if (this.test_settings.ie_driver) {
    deprecationNotice('ie_driver', 'webdriver.ie.driver');
    this.settings.selenium.cli_args['webdriver.ie.driver'] = this.test_settings.ie_driver;
  }

  return this;
}

/**
 * Merge current environment specific cli_args into the main cli_args object to be passed to selenium
 *
 * @returns {CliRunner}
 */
 function mergeCliArgs() {
  if (Utils.isObject(this.test_settings.cli_args)) {
    for (var prop in this.test_settings.cli_args) {
      if (this.test_settings.cli_args.hasOwnProperty(prop)) {
        this.settings.selenium.cli_args[prop] = this.test_settings.cli_args[prop];
      }
    }
  }
  return this;
}